
from __future__ import annotations

import asyncio
import dataclasses
import os.path
import uuid
from argparse import ArgumentParser, Namespace
from collections.abc import Set, Mapping, AsyncIterator
from contextlib import closing, asynccontextmanager, AsyncExitStack
from datetime import datetime, timezone
from importlib.resources import files
from secrets import token_bytes
from typing import Any, Final

from pysasl.creds.server import ServerCredentials

from pymap.config import BackendCapability, IMAPConfig
from pymap.exceptions import AuthorizationFailure, InvalidAuth, \
    NotAllowedError, UserNotFound
from pymap.health import HealthStatus
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import LoginInterface, IdentityInterface
from pymap.interfaces.token import TokenCredentials
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials.flag import Flag, Recent
from pymap.token import AllTokens
from pymap.user import Passwords, UserMetadata

from .filter import FilterSet
from .mailbox import Message, MailboxData, MailboxSet
from ..session import BaseSession

__all__ = ['DictBackend', 'Config']


class DictBackend(BackendInterface):
    """Defines a backend that uses an in-memory dictionary for example usage
    and integration testing.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._status = HealthStatus()

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @property
    def status(self) -> HealthStatus:
        return self._status

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser: ArgumentParser = subparsers.add_parser(
            name, help='in-memory backend')
        parser.add_argument('--demo-data', action='store_true',
                            help='load initial demo data')
        parser.add_argument('--demo-user', default='demouser',
                            metavar='VAL', help='demo user ID')
        parser.add_argument('--demo-password', default='demopass',
                            metavar='VAL', help='demo user password')
        return parser

    @classmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[DictBackend, Config]:
        config = Config.from_args(args, **overrides)
        login = Login(config)
        await cls._add_demo_user(config, login)
        return cls(login, config), config

    @classmethod
    async def _add_demo_user(cls, config: Config, login: Login) -> None:
        hashed_password = await Passwords(config).hash_password(
            config.demo_password)
        demo_user = UserMetadata(config, config.demo_user,
                                 password=hashed_password)
        await login.demo_user_identity.set(demo_user)

    async def start(self, stack: AsyncExitStack) -> None:
        pass


class Config(IMAPConfig):
    """The config implementation for the dict backend."""

    def __init__(self, args: Namespace, *, demo_data: bool,
                 demo_user: str, demo_password: str,
                 demo_data_resource: str = __name__,
                 admin_key: bytes | None = None, **extra: Any) -> None:
        admin_key = admin_key or token_bytes()
        super().__init__(args, admin_key=admin_key, **extra)
        self._demo_data = demo_data
        self._demo_user = demo_user
        self._demo_password = demo_password
        self._demo_data_resource = demo_data_resource
        self.set_cache: dict[str, tuple[MailboxSet, FilterSet]] = {}

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def demo_data(self) -> bool:
        """True if demo data should be loaded at startup."""
        return self._demo_data

    @property
    def demo_data_resource(self) -> str:
        """Resource path of demo data files."""
        return self._demo_data_resource

    @property
    def demo_user(self) -> str:
        """A login name that is valid at startup, which defaults to
        ``demouser``.

        """
        return self._demo_user

    @property
    def demo_password(self) -> str:
        """The password for the :attr:`.demo_user` login name, which defaults
        to ``demopass``.

        """
        return self._demo_password

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {**super().parse_args(args),
                'demo_data': args.demo_data,
                'demo_user': args.demo_user,
                'demo_password': args.demo_password}


class Session(BaseSession[Message]):
    """The session implementation for the dict backend."""

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the dict backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config: Final = config
        self.users_dict: dict[str, UserMetadata] = {}
        self.tokens_dict: dict[str, tuple[str, bytes]] = {}
        self._passwords = Passwords(config)
        self._tokens = AllTokens(config)

    @property
    def tokens(self) -> AllTokens:
        return self._tokens

    @property
    def demo_user_identity(self) -> Identity:
        return Identity(self.config.demo_user, self, None, frozenset())

    async def authenticate(self, credentials: ServerCredentials) -> Identity:
        authcid = credentials.authcid
        roles: set[str] = set()
        token_id: str | None = None
        if isinstance(credentials, TokenCredentials):
            token_id = credentials.identifier
            roles.update(credentials.roles)
        identity = Identity(authcid, self, token_id, roles)
        try:
            user = await identity.get()
        except UserNotFound:
            await asyncio.sleep(self.config.invalid_user_sleep)
            user = UserMetadata(self.config, authcid)
        if not await self._passwords.check_password(user, credentials):
            raise InvalidAuth()
        roles |= user.roles
        return identity

    async def authorize(self, authenticated: IdentityInterface, authzid: str) \
            -> Identity:
        authcid = authenticated.name
        roles = authenticated.roles
        if authcid != authzid and 'admin' not in roles:
            raise AuthorizationFailure()
        return Identity(authzid, self, None, roles)


class Identity(IdentityInterface):
    """The identity implementation for the dict backend."""

    def __init__(self, name: str, login: Login, token_id: str | None,
                 roles: Set[str]) -> None:
        super().__init__()
        self.login: Final = login
        self.config: Final = login.config
        self._name = name
        self._roles = roles
        self._token_id = token_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def roles(self) -> frozenset[str]:
        return frozenset(self._roles)

    async def new_token(self, *, expiration: datetime | None = None) \
            -> str | None:
        token_id = uuid.uuid4().hex
        token_key = token_bytes()
        self.login.tokens_dict[token_id] = (self.name, token_key)
        return self.login.tokens.get_login_token(
            token_id, self.name, token_key, expiration=expiration)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        identity = self.name
        config = self.config
        _ = await self.get()
        mailbox_set, filter_set = config.set_cache.get(identity, (None, None))
        if not mailbox_set or not filter_set:
            mailbox_set = MailboxSet()
            filter_set = FilterSet()
            if config.demo_data and identity == config.demo_user:
                await self._load_demo(config.demo_data_resource,
                                      mailbox_set, filter_set)
            config.set_cache[identity] = (mailbox_set, filter_set)
        yield Session(identity, config, mailbox_set, filter_set)

    async def _load_demo(self, resource: str, mailbox_set: MailboxSet,
                         filter_set: FilterSet) -> None:
        inbox = await mailbox_set.get_mailbox('INBOX')
        await self._load_demo_mailbox(resource, 'INBOX', inbox)
        mbx_names = sorted(f.name
                           for f in files(resource).joinpath('demo').iterdir()
                           if f.is_dir())
        await self._load_demo_sieve(resource, filter_set)
        for name in mbx_names:
            if name != 'INBOX':
                await mailbox_set.add_mailbox(name)
                mbx = await mailbox_set.get_mailbox(name)
                await self._load_demo_mailbox(resource, name, mbx)

    async def _load_demo_sieve(self, resource: str,
                               filter_set: FilterSet) -> None:
        path = os.path.join('demo', 'sieve')
        sieve = files(resource).joinpath(path).read_bytes()
        await filter_set.put('demo', sieve)
        await filter_set.set_active('demo')

    async def _load_demo_mailbox(self, resource: str, name: str,
                                 mbx: MailboxData) -> None:
        path = os.path.join('demo', name)
        msg_names = sorted(f.name
                           for f in files(resource).joinpath(path).iterdir()
                           if f.is_file())
        for msg_name in msg_names:
            if msg_name == '.readonly':
                mbx._readonly = True
                continue
            elif msg_name.startswith('.'):
                continue
            msg_path = os.path.join(path, msg_name)
            with closing(files(resource).joinpath(msg_path).open('rb')) \
                    as msg_stream:
                flags_line = msg_stream.readline()
                msg_timestamp = float(msg_stream.readline())
                msg_data = msg_stream.read()
            msg_dt = datetime.fromtimestamp(msg_timestamp, timezone.utc)
            msg_flags = {Flag(flag) for flag in flags_line.split()}
            if Recent in msg_flags:
                msg_flags.remove(Recent)
                msg_recent = True
            else:
                msg_recent = False
            append_msg = AppendMessage(msg_data, msg_dt, frozenset(msg_flags))
            await mbx.append(append_msg, recent=msg_recent)

    async def get(self) -> UserMetadata:
        token_id = self._token_id
        name = self.name
        token_key: bytes | None = None
        if token_id is not None:
            try:
                name, token_key = self.login.tokens_dict[token_id]
            except KeyError as exc:
                raise UserNotFound() from exc
            else:
                self._name = name
        user = self.login.users_dict.get(name)
        if user is None:
            raise UserNotFound(self.name)
        return dataclasses.replace(user, token_key=token_key)

    async def set(self, user: UserMetadata) -> None:
        if 'admin' not in self._roles and user.roles:
            raise NotAllowedError('Cannot assign roles.')
        self.login.users_dict[self.name] = user

    async def delete(self) -> None:
        try:
            del self.login.users_dict[self.name]
        except KeyError as exc:
            raise UserNotFound(self.name) from exc
        self.config.set_cache.pop(self.name, None)
        for token_id, (name, _) in list(self.login.tokens_dict.items()):
            if name == self.name:
                del self.login.tokens_dict[token_id]




from __future__ import annotations

import os
import os.path
from abc import abstractmethod, ABCMeta
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from tempfile import NamedTemporaryFile
from typing import TypeVar, Generic, Any, IO, Self

from pymap.concurrent import FileLock

__all__ = ['FileReadable', 'FileWriteable']

_RT = TypeVar('_RT', bound='FileReadable')
_WT = TypeVar('_WT', bound='FileWriteable')


class FileReadable(metaclass=ABCMeta):

    def __init__(self, path: str) -> None:
        super().__init__()
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    @classmethod
    def get_file(cls, path: str) -> str:
        return path

    @classmethod
    def get_lock(cls, path: str) -> str | None:
        return None

    @classmethod
    @abstractmethod
    def get_default(cls, path: str) -> Self:
        ...

    @classmethod
    @abstractmethod
    def open(cls, path: str, fp: IO[str]) -> Self:
        ...

    @abstractmethod
    def read(self, fp: IO[str]) -> None:
        ...

    @classmethod
    @asynccontextmanager
    async def _noop_lock(cls) -> AsyncIterator[None]:
        yield

    @classmethod
    def read_lock(cls, path: str) -> AbstractAsyncContextManager[None]:
        lock_path = cls.get_lock(path)
        if lock_path is not None:
            return FileLock(lock_path).read_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def write_lock(cls, path: str) -> AbstractAsyncContextManager[None]:
        lock_file = cls.get_lock(path)
        if lock_file is not None:
            return FileLock(lock_file).write_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def file_exists(cls, path: str) -> bool:
        file_path = cls.get_file(path)
        return os.path.exists(file_path)

    @classmethod
    def file_open(cls, path: str) -> Self:
        file_path = cls.get_file(path)
        try:
            with open(file_path, 'r') as in_file:
                return cls.open(path, in_file)
        except FileNotFoundError:
            return cls.get_default(path)

    @classmethod
    def file_read(cls, path: str) -> Self:
        file_path = cls.get_file(path)
        try:
            with open(file_path, 'r') as in_file:
                ret = cls.open(path, in_file)
                ret.read(in_file)
                return ret
        except FileNotFoundError:
            return cls.get_default(path)

    @classmethod
    def with_read(cls: type[_RT], path: str) \
            -> AbstractAsyncContextManager[_RT]:
        return _FileReadWith(path, cls)


class FileWriteable(FileReadable, metaclass=ABCMeta):

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self._touched = False
        self._watched = False

    @property
    @abstractmethod
    def empty(self) -> bool:
        ...

    @property
    def touched(self) -> bool:
        return self._touched

    def touch(self) -> None:
        if self._watched:
            self._touched = True

    @abstractmethod
    def write(self, fp: IO[str]) -> None:
        ...

    def file_delete(self) -> None:
        file_path = self.get_file(self.path)
        os.remove(file_path)
        self._touched = False

    def file_write(self) -> None:
        file_path = self.get_file(self.path)
        with NamedTemporaryFile('w', delete=False) as tmp:
            self.write(tmp)
        os.rename(tmp.name, file_path)
        self._touched = False

    @classmethod
    def with_init(cls: type[_WT], path: str) \
            -> AbstractAsyncContextManager[_WT]:
        return _FileInitWith(path, cls)

    @classmethod
    def with_write(cls: type[_WT], path: str) \
            -> AbstractAsyncContextManager[_WT]:
        return _FileWriteWith(path, cls)


class _FileReadWith(Generic[_RT]):

    def __init__(self, path: str, cls: type[_RT]) -> None:
        super().__init__()
        self._path = path
        self._cls = cls
        self._obj: _RT | None = None

    async def __aenter__(self) -> _RT:
        path = self._path
        cls = self._cls
        async with cls.read_lock(path):
            self._obj = obj = cls.file_read(path)
        return obj

    async def __aexit__(self, exc_type: Any, exc_val: Any,
                        exc_tb: Any) -> bool:
        return False


class _FileInitWith(Generic[_WT]):

    def __init__(self, path: str, cls: type[_WT]) -> None:
        super().__init__()
        self._path = path
        self._cls = cls
        self._obj: _WT | None = None

    async def __aenter__(self) -> _WT:
        path = self._path
        cls = self._cls
        if cls.file_exists(path):
            async with cls.read_lock(path):
                self._obj = obj = cls.file_open(path)
        else:
            async with cls.write_lock(path):
                self._obj = obj = cls.file_open(path)
                obj.file_write()
        return obj

    async def __aexit__(self, exc_type: Any, exc_val: Any,
                        exc_tb: Any) -> bool:
        return False


class _FileWriteWith(Generic[_WT]):

    def __init__(self, path: str, cls: type[_WT]) -> None:
        super().__init__()
        self._path = path
        self._cls = cls
        self._exists = False
        self._lock: AbstractAsyncContextManager[None] | None = None
        self._obj: _WT | None = None

    async def _acquire_lock(self) -> None:
        self._lock = self._cls.write_lock(self._path)
        await self._lock.__aenter__()

    async def _release_lock(self) -> None:
        lock = self._lock
        if not lock:
            raise RuntimeError()  # Lock not acquired.
        await lock.__aexit__(None, None, None)

    async def __aenter__(self) -> _WT:
        path = self._path
        cls = self._cls
        await self._acquire_lock()
        self._exists = cls.file_exists(path)
        self._obj = obj = cls.file_read(path)
        obj._watched = True
        return obj

    async def __aexit__(self, exc_type: Any, exc_val: Any,
                        exc_tb: Any) -> bool:
        try:
            obj = self._obj
            if obj is None:
                raise RuntimeError()  # Context manager not entered.
            obj._watched = False
            if not exc_type and obj.touched:
                if not obj.empty:
                    obj.file_write()
                elif self._exists:
                    obj.file_delete()
            return False
        finally:
            await self._release_lock()




from __future__ import annotations

from dataclasses import dataclass
from typing import overload, Final

from pysasl.creds.server import ServerCredentials
from pysasl.hashing import Cleartext
from pysasl.identity import Identity

from .config import IMAPConfig
from .frozen import frozendict

__all__ = ['Passwords', 'UserMetadata']

_empty_roles: frozenset[str] = frozenset()
_empty_params: frozendict[str, str] = frozendict()


class Passwords:
    """Helper utility for performing password operations in a CPU bound pool.

    Args:
        config: The IMAP config object.

    """

    __slots__ = ['password_prep', 'cpu_subsystem', 'hash_context']

    def __init__(self, config: IMAPConfig) -> None:
        super().__init__()
        self.password_prep: Final = config.password_prep
        self.cpu_subsystem: Final = config.cpu_subsystem
        self.hash_context: Final = config.hash_context

    @overload
    async def hash_password(self, password: str) -> str:
        ...

    @overload
    async def hash_password(self, password: None) -> None:
        ...

    async def hash_password(self, password: str | None) -> str | None:
        """Hash the given *password* using the configured hash algorithm.

        Args:
            password: The password string to hash.

        """
        if password is not None:
            fut = self._hash_password(password)
            return await self.cpu_subsystem.execute(fut)
        return None

    async def _hash_password(self, password: str) -> str:
        prepped = self.password_prep(password)
        return self.hash_context.hash(prepped)

    async def check_password(self, identity: Identity,
                             credentials: ServerCredentials) -> bool:
        """Check the *credentials* against the SASL *identity*.

        Args:
            identity: The identity to check against.
            credentials: The credentials to check with.

        """
        fut = self._check_password(identity, credentials)
        return await self.cpu_subsystem.execute(fut)

    async def _check_password(self, identity: Identity,
                              credentials: ServerCredentials) -> bool:
        return credentials.verify(identity)


@dataclass(frozen=True)
class UserMetadata(Identity):
    """Defines the attributes associated with a user identity."""

    #: The config object.
    config: IMAPConfig

    #: The user identity name.
    name: str

    #: The password string or hash digest.
    password: str | None = None

    #: The private key used to verify tokens.
    token_key: bytes | None = None

    #: The set of roles assigned to the user.
    roles: frozenset[str] = _empty_roles

    #: Additional parameters associated with the user.
    params: frozendict[str, str] = _empty_params

    @property
    def authcid(self) -> str:
        return self.name

    def compare_secret(self, value: str) -> bool:
        password = self.password
        if password is not None:
            hash_context = self.config.hash_context.copy()
            return hash_context.verify(value, password)
        return False

    def get_clear_secret(self) -> str | None:
        if isinstance(self.config.hash_context, Cleartext):
            return self.password
        return None



"""Frozen collection types not (yet) provided by the Python library."""

from __future__ import annotations

from collections.abc import Hashable, Iterable, Mapping, Sequence
from typing import final, Any, TypeAlias, TypeVar

__all__ = ['HashableT', 'ValueT', 'FrozenDict', 'FrozenList',
           'frozendict', 'frozenlist']

#: The type variable representing a mapping key.
HashableT = TypeVar('HashableT', bound='Hashable')

#: The type variable representing a value in a collection.
ValueT = TypeVar('ValueT')


@final
class FrozenDict(Mapping[HashableT, ValueT], Hashable):
    """A :class:`~collections.abc.Mapping` that does not support mutation.

    Args:
        mapping: Initiailize with the contents of another mapping.

    """

    __slots__ = ['_mapping', '_hash']

    _empty: Mapping[Any, Any] = {}

    def __init__(self, /, mapping: Mapping[HashableT, ValueT] |
                 Iterable[tuple[HashableT, ValueT]] = _empty) -> None:
        super().__init__()
        self._mapping: dict[HashableT, ValueT] = dict(mapping)
        self._hash: int | None = None

    def __getitem__(self, key: Any) -> Any:
        return self._mapping.__getitem__(key)

    def __iter__(self) -> Any:
        return self._mapping.__iter__()

    def __len__(self) -> Any:
        return self._mapping.__len__()

    def __contains__(self, key: Any) -> Any:
        return self._mapping.__contains__(key)

    def keys(self) -> Any:
        return self._mapping.keys()

    def values(self) -> Any:
        return self._mapping.values()

    def items(self) -> Any:
        return self._mapping.items()

    def get(self, key: Any, *args: Any, **kwargs: Any) -> Any:
        return self._mapping.get(key, *args, **kwargs)

    def __eq__(self, other: Any) -> Any:
        return self._mapping.__eq__(other)

    def __hash__(self) -> Any:
        saved = self._hash
        if saved is None:
            self._hashed = saved = hash(tuple(self.items()))
        return saved


@final
class FrozenList(Sequence[ValueT], Hashable):
    """A :class:`~collections.abc.Sequence` that does not support mutation.

    Args:
        iterable: Initiailize with the contents of another iterable.

    """

    __slots__ = ['_sequence']

    _empty: Sequence[Any] = []

    def __init__(self, /, iterable: Iterable[ValueT] = _empty) -> None:
        super().__init__()
        self._sequence: tuple[ValueT, ...] = tuple(list(*iterable))

    def __getitem__(self, index: Any) -> Any:
        return self._sequence.__getitem__(index)

    def __len__(self) -> Any:
        return self._sequence.__len__()

    def __contains__(self, element: Any) -> Any:
        return self._sequence.__contains__(element)

    def __iter__(self) -> Any:
        return self._sequence.__iter__()

    def __reversed__(self) -> Any:
        return self._sequence.__reversed__()

    def index(self, value: Any, *args: Any, **kwargs: Any) -> Any:
        return self._sequence.index(value, *args, **kwargs)

    def count(self, value: Any) -> Any:
        return self._sequence.count(value)

    def __eq__(self, other: Any) -> Any:
        return self._sequence.__eq__(other)

    def __hash__(self) -> Any:
        return self._sequence.__hash__()


#: An alias for :class:`FrozenDict`.
frozendict: TypeAlias = FrozenDict

#: An alias for :class:`FrozenList`.
frozenlist: TypeAlias = FrozenList




from __future__ import annotations

import asyncio
import os.path
import secrets
import uuid
from argparse import Action, ArgumentParser, Namespace
from collections.abc import Mapping, AsyncIterator, Set
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager, AsyncExitStack
from datetime import datetime
from typing import Any, Final

from pysasl.creds.server import ServerCredentials

from pymap.concurrent import Subsystem
from pymap.config import BackendCapability, IMAPConfig
from pymap.exceptions import AuthorizationFailure, InvalidAuth, \
    NotAllowedError, UserNotFound
from pymap.filter import PluginFilterSet, SingleFilterSet
from pymap.frozen import frozendict
from pymap.health import HealthStatus
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import LoginInterface, IdentityInterface
from pymap.interfaces.token import TokenCredentials, TokensInterface
from pymap.token import AllTokens
from pymap.user import Passwords, UserMetadata

from .layout import MaildirLayout
from .mailbox import Message, Maildir, MailboxSet
from .users import UsersFile, PasswordsFile, TokensFile, GroupsFile
from ..session import BaseSession

__all__ = ['MaildirBackend', 'Config']


class MaildirBackend(BackendInterface):
    """Defines an on-disk backend that uses :class:`~mailbox.Maildir` for
    mailbox storage and `MailboxFormat/Maildir
    <https://wiki2.dovecot.org/MailboxFormat/Maildir>`_ for metadata storage.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._status = HealthStatus()

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @property
    def status(self) -> HealthStatus:
        return self._status

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser: ArgumentParser = subparsers.add_parser(
            name, help='on-disk backend')
        parser.add_argument('base_dir', metavar='DIR', action=_BaseDirAction,
                            help='base directory for mailbox relative paths')
        parser.add_argument('--concurrency', metavar='NUM', type=int,
                            help='maximum number of IO workers')
        parser.add_argument('--layout', metavar='TYPE', default='++',
                            help='maildir directory layout')
        parser.add_argument('--colon', metavar='CHAR', default=None,
                            help='info delimiter in mail filename')
        return parser

    @classmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[MaildirBackend, Config]:
        config = Config.from_args(args)
        login = Login(config)
        if not os.path.exists(config.base_dir):
            os.mkdir(config.base_dir)
        return cls(login, config), config

    async def start(self, stack: AsyncExitStack) -> None:
        pass


class _BaseDirAction(Action):

    def __call__(self, parser: ArgumentParser, namespace: Namespace,
                 values: Any, option_string: str | None = None) -> None:
        assert isinstance(values, str)
        if os.path.isfile(values):
            raise parser.error(
                f'{self.metavar} argument {values!r} must not be a file')
        setattr(namespace, self.dest, values)


class Config(IMAPConfig):
    """The config implementation for the maildir backend.

    Args:
        args: The command-line arguments.
        base_dir: The base directory for all relative mailbox paths.
        layout: The Maildir directory layout.
        colon: The info delimiter in mail filename.
        hash_interface: The hash algorithm to use for passwords.

    """

    def __init__(self, args: Namespace, *, base_dir: str,
                 layout: str, colon: str | None,
                 **extra: Any) -> None:
        super().__init__(args, admin_key=secrets.token_bytes(), **extra)
        self._base_dir = base_dir
        self._layout = layout
        self._colon = colon

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def base_dir(self) -> str:
        """The base directory for all relative paths."""
        return self._base_dir

    @property
    def layout(self) -> str:
        """The Maildir directory layout name.

        See Also:
            :class:`~pymap.backend.maildir.layout.MaildirLayout`

        """
        return self._layout

    @property
    def colon(self) -> str | None:
        """The info delimiter in mail filename.

        See Also:
            Note on ``colon`` in :class:`mailbox.Maildir`.

        """
        return self._colon

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        executor = ThreadPoolExecutor(args.concurrency)
        subsystem = Subsystem.for_executor(executor)
        return {**super().parse_args(args),
                'base_dir': args.base_dir,
                'layout': args.layout,
                'colon': args.colon,
                'subsystem': subsystem}


class FilterSet(PluginFilterSet[bytes], SingleFilterSet[bytes]):

    def __init__(self, user_dir: str) -> None:
        super().__init__('sieve', bytes)
        self._user_dir = user_dir

    async def replace_active(self, value: bytes | None) -> None:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        if value is None:
            try:
                os.unlink(path)
            except FileNotFoundError:
                pass
        else:
            with open(path, 'wb') as sieve_file:
                sieve_file.write(value)

    async def get_active(self) -> bytes | None:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        try:
            with open(path, 'rb') as sieve_file:
                return sieve_file.read()
        except FileNotFoundError:
            return None


class Session(BaseSession[Message]):
    """The session implementation for the maildir backend."""

    resource = __name__

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the maildir backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config
        self._passwords = Passwords(config)
        self._tokens = AllTokens(config)

    @property
    def tokens(self) -> AllTokens:
        return self._tokens

    async def authenticate(self, credentials: ServerCredentials) \
            -> Identity:
        config = self.config
        authcid = credentials.authcid
        roles: set[str] = set()
        token_id: str | None = None
        if isinstance(credentials, TokenCredentials):
            token_id = credentials.identifier
            roles.update(credentials.roles)
        identity = Identity(config, self.tokens, authcid, token_id, roles)
        try:
            user = await identity.get()
        except UserNotFound:
            await asyncio.sleep(self.config.invalid_user_sleep)
            user = UserMetadata(config, authcid)
        roles |= user.roles
        if not await self._passwords.check_password(user, credentials):
            raise InvalidAuth()
        return identity

    async def authorize(self, authenticated: IdentityInterface, authzid: str) \
            -> Identity:
        authcid = authenticated.name
        roles = authenticated.roles
        if authcid != authzid and roles.isdisjoint({'sudo', 'admin'}):
            raise AuthorizationFailure()
        return Identity(self.config, self.tokens, authzid, None, roles)


class Identity(IdentityInterface):
    """The identity implementation for the maildir backend."""

    def __init__(self, config: Config, tokens: TokensInterface,
                 name: str, token_id: str | None, roles: Set[str]) -> None:
        super().__init__()
        self.config: Final = config
        self.tokens: Final = tokens
        self._name = name
        self._token_id = token_id
        self._roles = frozenset(roles)
        self._base_dir = config.base_dir

    @property
    def name(self) -> str:
        return self._name

    @property
    def roles(self) -> frozenset[str]:
        return self._roles

    async def new_token(self, *, expiration: datetime | None = None) \
            -> str | None:
        base_dir = self._base_dir
        name = self.name
        identifier = uuid.uuid4().hex
        key = secrets.token_bytes()
        async with UsersFile.with_write(base_dir) as users_file, \
                TokensFile.with_write(base_dir) as tokens_file:
            if not users_file.has(name):
                raise UserNotFound(name)
            tokens_file.set(tokens_file.build_record(
                identifier, key, name))
        return self.tokens.get_login_token(identifier, name, key)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        config = self.config
        base_dir = self._base_dir
        name = self.name
        async with UsersFile.with_read(base_dir) as users_file:
            try:
                user_record = users_file.get(name)
            except KeyError as exc:
                raise UserNotFound(name) from exc
            else:
                mailbox_path = user_record.home_dir
        maildir, layout = self._load_maildir(mailbox_path)
        mailbox_set = MailboxSet(maildir, layout)
        filter_set = FilterSet(layout.path)
        yield Session(self.name, config, mailbox_set, filter_set)

    def _load_maildir(self, mailbox_path: str) \
            -> tuple[Maildir, MaildirLayout[Any]]:
        full_path = os.path.join(self._base_dir, mailbox_path)
        layout = MaildirLayout.get(full_path, self.config.layout, Maildir)
        create = not os.path.exists(full_path)
        maildir = Maildir(full_path, create=create)
        colon = self.config.colon
        if colon is not None:
            maildir.colon = colon
        return maildir, layout

    async def get(self) -> UserMetadata:
        base_dir = self._base_dir
        name = self.name
        token_id = self._token_id
        password: str | None = None
        token_key: bytes | None = None
        roles = set(self._roles)
        async with UsersFile.with_read(base_dir) as users_file, \
                GroupsFile.with_read(base_dir) as groups_file:
            if token_id is None:
                async with PasswordsFile.with_read(base_dir) as passwords_file:
                    try:
                        password_record = passwords_file.get(name)
                    except KeyError:
                        pass
                    else:
                        password = password_record.password
                        if not password or password[0] in ('*', '!'):
                            password = None  # disabled
            else:
                async with TokensFile.with_read(base_dir) as tokens_file:
                    try:
                        token = tokens_file.get(token_id)
                    except KeyError as exc:
                        raise UserNotFound() from exc
                    else:
                        self._name = name = token.users_list
                        token_key = bytes.fromhex(token.password)
            try:
                user_record = users_file.get(name)
            except KeyError as exc:
                raise UserNotFound(name) from exc
            else:
                if user_record.uid == '0':
                    roles.add('admin')
                mailbox_path = user_record.home_dir
            for role in groups_file.get_user(name):
                roles.add(role.name)
        params = frozendict({'mailbox_path': mailbox_path})
        return UserMetadata(self.config, name, password, token_key,
                            frozenset(roles), params)

    async def set(self, metadata: UserMetadata) -> None:
        base_dir = self._base_dir
        name = self.name
        password = metadata.password if metadata.password is not None else '*'
        mailbox_path = metadata.params.get('mailbox_path', name)
        roles = metadata.roles
        if self.roles.isdisjoint({'sudo', 'admin'}):
            async with UsersFile.with_read(base_dir) as users_file, \
                    GroupsFile.with_read(base_dir) as groups_file:
                try:
                    existing_user = users_file.get(name)
                except KeyError:
                    pass
                else:
                    if mailbox_path != existing_user.home_dir:
                        raise NotAllowedError('Cannot assign parameters.')
                existing_roles = {group.name for group in
                                  groups_file.get_user(name)}
                if metadata.roles != existing_roles:
                    raise NotAllowedError('Cannot assign roles.')
        async with UsersFile.with_write(base_dir) as users_file, \
                PasswordsFile.with_write(base_dir) as passwords_file, \
                GroupsFile.with_write(base_dir) as groups_file:
            users_file.set(users_file.build_record(name, mailbox_path))
            passwords_file.set(passwords_file.build_record(name, password))
            groups_file.remove_user(name)
            groups_file.merge(groups_file.build_record(role, name)
                              for role in roles)

    async def delete(self) -> None:
        base_dir = self._base_dir
        name = self.name
        async with UsersFile.with_write(base_dir) as users_file, \
                PasswordsFile.with_write(base_dir) as passwords_file, \
                TokensFile.with_write(base_dir) as tokens_file, \
                GroupsFile.with_write(base_dir) as groups_file:
            try:
                users_file.remove(name)
            except KeyError as exc:
                raise UserNotFound(name) from exc
            try:
                passwords_file.remove(name)
            except KeyError:
                pass
            tokens_file.remove_user(name)
            groups_file.remove_user(name)




from __future__ import annotations

import os.path
import random
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import ClassVar, IO, Literal, Self

from pymap.mailbox import MailboxSnapshot

from .io import FileWriteable

__all__ = ['Record', 'UidList']


@dataclass(frozen=True)
class Record:
    """Defines a single record read from the UID list file.

    Args:
        uid: The message UID of the record.
        fields: The metadata fields of the record.
        filename: The filename of the record.

    """

    uid: int
    fields: Mapping[str, str]
    filename: str

    @property
    def key(self) -> str:
        """The :class:`~mailbox.Maildir` key value."""
        return self.filename.split(':', 1)[0]


class UidList(FileWriteable):
    """Maintains the file with UID mapping to maildir files.

    Args:
        path: The directory of the file.
        uid_validity: The UID validity value.
        next_uid: The next assignable message UID value.
        global_uid: The 128-bit global mailbox UID.

    """

    #: The UID list file name, stored in the mailbox directory.
    FILE_NAME: ClassVar[str] = 'dovecot-uidlist'

    #: The UID list lock file, stored adjacent to the UID list file.
    LOCK_FILE: ClassVar[str] = 'dovecot-uidlist.lock'

    def __init__(self, path: str, uid_validity: int,
                 next_uid: int, global_uid: bytes | None = None) -> None:
        super().__init__(path)
        self.uid_validity = uid_validity
        self.next_uid = next_uid
        self.global_uid = global_uid or self._create_guid()
        self._records: dict[int, Record] = {}

    @property
    def empty(self) -> Literal[False]:
        return False

    @property
    def records(self) -> Iterable[Record]:
        """The records contained in the UID list file."""
        return self._records.values()

    def get(self, uid: int) -> Record:
        """Get a single record by its UID.

        Args:
            uid: The message UID.

        Raises:
            KeyError: The UID does not exist.

        """
        return self._records[uid]

    def get_all(self, uids: Iterable[int]) -> Mapping[int, Record]:
        """Get records by a set of UIDs.

        Args:
            uids: The message UIDs.

        """
        return {uid: self._records[uid] for uid in uids
                if uid in self._records}

    def set(self, rec: Record) -> None:
        """Add or update the record in the UID list file."""
        self._records[rec.uid] = rec
        self.touch()

    def remove(self, uid: int) -> None:
        """Remove the record from the UID list file.

        Raises:
            KeyError: The UID does not exist.

        """
        del self._records[uid]
        self.touch()

    @classmethod
    def _build_line(cls, rec: Record) -> str:
        parts = ['%d' % rec.uid]
        for key, val in sorted(rec.fields.items()):
            if val is not None:
                parts.append(' ')
                parts.append(key[0:1])
                parts.append(val)
        parts.append(' :')
        parts.append(rec.filename)
        parts.append('\r\n')
        return ''.join(parts)

    @classmethod
    def _read_line(cls, line: str) -> Record:
        before, filename = line.split(':', 1)
        fields: dict[str, str] = {}
        data = before.split(' ')
        num = int(data[0])
        for col in data[1:]:
            if col:
                fields[col[0]] = col[1:]
        return Record(num, fields, filename.rstrip())

    @classmethod
    def _read_header(cls, path: str, line: str) -> Self:
        data = line.split()
        if data[0] != '3':
            raise ValueError(line)
        uid_validity: int | None = None
        next_uid: int | None = None
        global_uid: bytes | None = None
        for field in data[1:]:
            if field[0] == 'V':
                uid_validity = int(field[1:])
            elif field[0] == 'N':
                next_uid = int(field[1:])
            elif field[0] == 'G':
                global_uid = field[1:].encode('ascii')
        if uid_validity is None or next_uid is None or global_uid is None:
            raise ValueError(line)
        return cls(path, uid_validity, next_uid, global_uid)

    @classmethod
    def _create_guid(cls) -> bytes:
        return b'%032x' % random.getrandbits(128)

    def _build_header(self) -> str:
        global_uid = self.global_uid.decode('ascii')
        return ''.join(['3 V', str(self.uid_validity),
                        ' N', str(self.next_uid),
                        ' G', global_uid, '\r\n'])

    @classmethod
    def get_file(cls, path: str) -> str:
        return os.path.join(path, cls.FILE_NAME)

    @classmethod
    def get_lock(cls, path: str) -> str:
        return os.path.join(path, cls.LOCK_FILE)

    @classmethod
    def get_default(cls, path: str) -> Self:
        return cls(path, MailboxSnapshot.new_uid_validity(), 1)

    def write(self, fp: IO[str]) -> None:
        fp.write(self._build_header())
        for rec in self.records:
            fp.write(self._build_line(rec))

    @classmethod
    def open(cls, path: str, fp: IO[str]) -> Self:
        header = fp.readline()
        ret = cls._read_header(path, header)
        return ret

    def read(self, fp: IO[str]) -> None:
        for line in fp:
            self.set(self._read_line(line))
